#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['axuy']

package_data = \
{'': ['*'], 'axuy': ['shaders/*']}

install_requires = \
['appdirs', 'glfw>=1.8', 'moderngl', 'numpy', 'Pillow', 'pyrr', 'setuptools']

entry_points = \
{'console_scripts': ['axuy = axuy.__main__:main']}

setup(name='axuy',
      version='0.0.10',
      description='Axuy is a minimalist peer-to-peer first-person shooter.',
      author='Nguyễn Gia Phong',
      author_email='mcsinyx@disroot.org',
      url='https://github.com/McSinyx/axuy',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      entry_points=entry_points,
      python_requires='>=3.6',
     )
